use std::{collections::HashMap, io};

use rayon::prelude::*;

use crate::{
    chunk::{read_chunks_messages, ChunkHeader},
    index::{BagIndex, IndexData},
    reader::BagReader,
    Result,
};

#[derive(Default, Debug)]
pub struct BagInfo {
    pub total_uncompressed: u64,
    pub per_connection: HashMap<u32, u64>,
}

impl BagInfo {
    fn combine(mut self, other: BagInfo) -> BagInfo {
        self.total_uncompressed += other.total_uncompressed;
        for (conn, count) in other.per_connection {
            *self.per_connection.entry(conn).or_insert(0) += count;
        }
        self
    }

    pub fn compute<R>(reader: R, index: &BagIndex) -> Result<BagInfo>
    where
        R: BagReader + io::Seek + Clone + Send + Sync,
    {
        index
            .chunks
            .par_iter()
            .try_fold(BagInfo::default, move |mut info, chunk| -> Result<_> {
                let mut reader = reader.clone();
                reader.seek(io::SeekFrom::Start(chunk.pos))?;
                let chunk_header = ChunkHeader::read(&mut reader)?;
                info.total_uncompressed += chunk_header.uncompressed_size as u64;
                reader.skip_data()?;
                for _ in &chunk.connections {
                    let index = IndexData::read(&mut reader)?;
                    *info.per_connection.entry(index.conn_id).or_insert(0) +=
                        index.entries.len() as u64;
                }
                Ok(info)
            })
            .try_reduce(BagInfo::default, |a, b| Ok(a.combine(b)))
    }

    pub fn compute_without_index<R>(reader: R, index: &BagIndex) -> Result<BagInfo>
    where
        R: BagReader + io::Seek + Clone + Send,
    {
        read_chunks_messages(reader, &index.chunks)
            .try_fold(BagInfo::default, |mut info, chunk| -> Result<_> {
                let data = chunk?;
                info.total_uncompressed += data.data.len() as u64;
                let count = info.per_connection.entry(data.header.conn_id).or_insert(0);
                *count += 1;
                Ok(info)
            })
            .try_reduce(BagInfo::default, |a, b| Ok(a.combine(b)))
    }
}
